const { exec } = require('child_process')
const fs = require('fs')
const path = require('path')
const { allowedFiles } = require('./filterUtils')
const getNodeModulesPath = require('../../utils/getNodeModulesPath')
const { getLibraryInstalledLocation } = require('../../utils/getLibraryInstalledLocation')
const { filterFiles } = require('../../utils/filterFiles')
const { Logger } = require('../../utils/Logger')
const { checkIfPluginsAreInstalled } = require('../../utils/PluginsInstallationUtils/checkIfPluginsAreInstalled')
const libraryInstalledLocation = getLibraryInstalledLocation()
const nodeModulesPathOfProject = getNodeModulesPath()

async function isMergeCommit() {
    return new Promise((resolve, reject) => {
        exec('git rev-parse -q --verify MERGE_HEAD', (error,stderr,stdout) => {
            if(error){
                reject(error.toString().trim())
            }
            else if(stderr){
                resolve(stderr.trim())
            }
            else if(stdout){
                resolve(stdout.trim())
            }
        })
    })

}

async function getCurrentBranch(){
    return new Promise((resolve, reject) => {
        exec('git rev-parse --abbrev-ref HEAD', (error,stderr,stdout) => {
            if(error){
                reject('Error fetching current branch')
            }
            else if(stderr){
                resolve(stderr)
            }
        })
    })
}

async function getStagedFiles(){
    return new Promise((resolve, reject) => {
        exec("git diff --staged --name-only",(error,stderr,stdout) =>{
           if (error){
            if(error != null)
            reject("Couldn't fetch staged files")
           }
           else if(stderr){
            resolve(stderr.trim().split('\n'))
           }
           else if(stdout.trim() === ''){
            resolve(stdout.trim())
           }
        })
    })
    
}

function getEslintExecutablePath(alternativeBinPath,libraryInstalledLocation){
    let eslintExecutablePath =  path.join(libraryInstalledLocation,'node_modules','eslint','bin','eslint.js')
    alternativeBinPath = path.join(nodeModulesPathOfProject,'node_modules','eslint','bin','eslint.js')
    let isEslintExecutablePresent = fs.existsSync(eslintExecutablePath) ? true : fs.existsSync(alternativeBinPath) ? true : false
    return {
        eslintExecutablePath,
        isEslintExecutablePresent
    }
}

async function findEslintErrors(file){
    let nodeModulesPathOfProject = getNodeModulesPath()
    let eslintConfigurationFilePath = path.resolve(nodeModulesPathOfProject,'.eslintrc.js')
    let isNodeModulesPresent = fs.existsSync(nodeModulesPathOfProject)
    const {eslintExecutablePath, isEslintExecutablePresent} = getEslintExecutablePath(nodeModulesPathOfProject,libraryInstalledLocation)
    
    if(isNodeModulesPresent){
        if(isEslintExecutablePresent){
            return new Promise ((resolve, reject) => {
                exec(`npx --ignore-existing "${eslintExecutablePath}" --config "${eslintConfigurationFilePath}" --no-inline-config --resolve-plugins-relative-to="${nodeModulesPathOfProject}/node_modules" ${file}`,(error,stderr,stdout) => {
                    if(stderr){
                        resolve(stderr.trim().split('\n'))
                    }
                    else if(error){
                        Logger.log(Logger.FAILURE_TYPE,error)
                        reject("Error executing eslint command")
                    }
                    else{
                        resolve([])
                    }
                })
            })
        }
        else{
            Logger.log(Logger.INFO_TYPE,'Eslint executable not found. make sure eslint plugin is installed')
        }
    }
    else{
        Logger.log(Logger.INFO_TYPE,'node_modules not found')
    }
}
async function calculateGitDiffForFile(branch_name,file){
    let gitDiffCommand = `git diff -U0 ${branch_name.trim()} ${file}`
    return new Promise ((resolve,reject) => {        
        exec(gitDiffCommand,(error, stderr) => {
            if(stderr){
                resolve(stderr.trim().split('\n'))
            }
            else if(error) {
                reject(error)
            }
        })
    })
}
function areAllPluginsInstalled(){
    let unInstalledPlugins = checkIfPluginsAreInstalled().uninstalledPlugins
    return unInstalledPlugins.length === 0 ? true : false
}
async function preCommitHook() {
    Logger.log(Logger.INFO_TYPE,'Executing pre commit hook...')
    Logger.log(Logger.INFO_TYPE,`working dir  : ${process.cwd()}`)
    try{
        let isMerge = await isMergeCommit();
        Logger.log(Logger.INFO_TYPE,'Looks like you have merged. So skipping pre commit check');
        process.exit(0);
    }
    catch(error){
        if(areAllPluginsInstalled()){
            let staged_files = []
            let eslintConfigFiles = ['lint.config.js']
            let exemptionFiles = []
            let current_branch = ''
            let hasEslintErrorsInChangedLines = false
            let areFilesStaged = false
            try{
                current_branch = await getCurrentBranch()
            }
            catch{
                Logger.log(Logger.INFO_TYPE,"Error fetching current branch")
            }
            try{
                staged_files = await getStagedFiles()
                if(!staged_files.length == 0){
                    staged_files = filterFiles(staged_files,eslintConfigFiles,true)
        
                    // staged_files = filterFiles(staged_files,exemptionFiles) //this is the code for giving exemption to a file during pre commit
                    // if(staged_files.length === 0){
                    //     Logger.log(Logger.SUCCESS_TYPE,`Commit Successful`)
                    //     process.exit(0)
                    // }
                    areFilesStaged = true
                    for (file in staged_files){
                        let currentFileName = staged_files[file]
                        let changedLinesArray = []
                        let eslintErrorsInChangedLines = []

                        if(path.extname(staged_files[file]) === '.js'){
                            try{
                                let eslintErrorsInFile= await findEslintErrors(staged_files[file])
                                if(staged_files[file] && typeof(staged_files[file]) == 'string'){
                                    if(! eslintErrorsInFile.length == 0){
                                        //git diff is computed and stored in an array
                                        let git_diff = await calculateGitDiffForFile(current_branch,staged_files[file])
                                        changedLinesArray = git_diff.filter((line) => line.startsWith('@@'))
                                        
                                        let changedLinesStartArray = []
                                        let changedLinesEndArray = []
                                        
                                        //Calculating changed lines in a file and storing them in respective arrays
                                        for (number of changedLinesArray){
                                            let changesStartLine = parseInt(number.split(' ')[2].split(',')[0])
                                            changedLinesStartArray.push(changesStartLine)
                                            let changesEndLine = number.split(' ')[2].split(',')[1]
                                            if(changesEndLine === undefined){
                                                changedLinesEndArray.push(changesStartLine)
                                            }
                                            else{
                                                changedLinesEndArray.push(changesStartLine + parseInt(changesEndLine) - 1)
                                            }
                                        }
                                        for (let error=1; error < eslintErrorsInFile.length - 2; error++){
                                            //eslintErrorsInFile[error].trim() - 69:26  error => Do not hardcode content. Use I18N key instead  no-hardcoding/no-hardcoding, 
                                            //eslintErrorsInFile[error].trim().split(' ')[0] => 69:26
                                            //eslintErrorsInFile[error].trim().split(' ')[0].split(':')[0] => 69
                                            
                                            let eslintErrorLineNumber = eslintErrorsInFile[error].trim().split(' ')[0].split(':')[0]
            
                                            for(lineNumber in changedLinesStartArray){
                                                if(eslintErrorLineNumber >= changedLinesStartArray[lineNumber] && 
                                                    eslintErrorLineNumber <= changedLinesEndArray[lineNumber]){
                                                        eslintErrorsInChangedLines.push(eslintErrorsInFile[error])
                                                    }
                                            }
                                        }
                                        if(eslintErrorsInChangedLines.length > 0){
                                            Logger.log(Logger.FAILURE_TYPE,`\x1b[1m${currentFileName}\x1b[0m`)
                                            for(eslintError of eslintErrorsInChangedLines){
                                                Logger.log(Logger.FAILURE_TYPE,`\x1b[37m${eslintError.trimEnd()}\x1b[0m`)
                                            }
                                            hasEslintErrorsInChangedLines = true
                                        }  
                                    }
                                } 
                            }
                            catch(err){
                                Logger.log(Logger.FAILURE_TYPE,err)
                                Logger.log(Logger.FAILURE_TYPE,"Error in executing eslint command")
                            }
                            
                        }
                    }
                }
                else if(staged_files.length === 0){
                    Logger.log(Logger.INFO_TYPE,'No files have been staged. Stage your files before committing')
                }
            }
            catch{
                Logger.log(Logger.INFO_TYPE,'Error executing pre commit hook')
            }
            if(hasEslintErrorsInChangedLines){
                Logger.log(Logger.FAILURE_TYPE,`There are eslint errors present. So commit is aborted`)
                process.exit(1)
            }
            else if(!hasEslintErrorsInChangedLines && areFilesStaged){
                Logger.log(Logger.SUCCESS_TYPE,`Commit Successful`)
                process.exit(0)
            }
        }
        else{
            Logger.log(Logger.FAILURE_TYPE,'Commit failed since some eslint plugins are not installed')
            Logger.log(Logger.INFO_TYPE,`Kindly execute the command \x1b[37mnpx ZDPrecommit setupPlugins \x1b[33mfrom the location where package.json is present to install the plugins`)
            Logger.log(Logger.INFO_TYPE,'Execute the command and kindly try committing again.')
            process.exit(1)
        }
    }
}


module.exports = {
    preCommitHook
}